/*global window: true*/
(function (GCN) {

	'use strict';

	/**
	 * @private
	 * @const
	 * @type {object<string, function(string, *)|string>} This hash is used to
	 *                                                    determine what
	 *                                                    property can be read
	 *                                                    for a part of a given
	 *                                                    type. *SELECT and
	 *                                                    PAGE parttypes are
	 *                                                    handled specially.
	 */
	var PART_TYPES = {};

	PART_TYPES = {
		STRING   : 'stringValue',
		RICHTEXT : 'stringValue',
		BOOLEAN  : 'booleanValue',
		IMAGE    : 'imageId',
		// (URL) file is the same as File (upload).
		FILE     : 'fileId',
		// (URL) folder is the same as Folder (upload).
		FOLDER   : 'folderId',
		OVERVIEW : 'overview',
		PAGE     : function (part, value) {
			if (jQuery.type(value) === 'number') {
				part.pageId = value;
				delete part.stringValue;
				return value;
			}

			if (typeof value !== 'undefined') {
				part.stringValue = value.toString();
				delete part.pageId;
				return value;
			}

			return part[jQuery.type(part.stringValue) === 'string'
				? 'stringValue' : 'pageId'];
		},

		SELECT : function (part, value) {
			if (value) {
				if (typeof value.datasourceId !== 'undefinded') {
					part.datasourceId = value.datasourceId;
				}

				if (value.options) {
					part.options = value.options;
				}

				if (value.selectedOptions) {
					part.selectedOptions = value.selectedOptions;
				}
			}

			return {
				datasourceId    : part.datasourceId,
				options         : part.options,
				selectedOptions : part.selectedOptions
			};
		},

		MULTISELECT : function (part, value) {
			return PART_TYPES.SELECT(part, value);
		},

		TEMPLATETAG : function (part, value) {
			if (value) {
				if (typeof value.templateId !== 'undefined') {
					part.templateId = value.templateId;
				}

				if (typeof value.templateTagId !== 'undefined') {
					part.templateTagId = value.templateTagId;
				}
			}

			return {
				templateId    : part.templateId,
				templateTagId : part.templateTagId
			};
		},

		PAGETAG : function (part, value) {
			if (value) {
				if (typeof value.pageId !== 'undefined') {
					part.pageId = value.pageId;
				}

				if (typeof value.pageTagId !== 'undefined') {
					part.pageTagId = value.pageTagId;
				}
			}

			return {
				templateId    : part.pageId,
				templateTagId : part.pageTagId
			};
		}
	};

	/**
	 * Determine the key inside tag properties where our value is stored, and
	 * retreive that value.
	 *
	 * @param {object} part The tag part we want to read.
	 * @return {*} The value that the given tag part holds.
	 * @throws CANNOT_READ_TAG_PART
	 */
	function getPartValue(part) {
		var propName = PART_TYPES[part.type];

		if (!propName) {
			GCN.error('CANNOT_READ_TAG_PART',
				'Cannot read or write to tag part', part);

			return null;
		}

		if (jQuery.type(propName) === 'function') {
			return propName(part);
		}

		return part[propName];
	}


	function setPartValue(part, value) {
		var prop = PART_TYPES[part.type];

		if (!prop) {
			GCN.error('CANNOT_READ_TAG_PART',
				'Cannot read or write to tag part', part);

			return null;
		}

		if (jQuery.type(prop) === 'function') {
			return prop(part, value);
		}

		part[prop] = value;

		return value;
	}

	/**
	 * Searches for the an Aloha editable object of the given id.
	 *
	 * @TODO: Once Aloha.getEditableById() is patched to not cause an
	 *        JavaScript exception if the element for the given ID is not found
	 *        then we can deprecate this function and use Aloha's instead.
	 *
	 * @static
	 * @param {string} id Id of Aloha.Editable object to find.
	 * @return {Aloha.Editable=} The editable object, if wound; otherwise null.
	 */
	function getAlohaEditableById(id) {
		var Aloha = (typeof window !== 'undefined') && window.Aloha;
		if (!Aloha) {
			return null;
		}

		// If the element is a textarea then route to the editable div.
		var element = jQuery('#' + id);
		if (element.length &&
				element[0].nodeName.toLowerCase() === 'textarea') {
			id += '-aloha';
		}

		var editables = Aloha.editables;
		var j = editables.length;
		while (j) {
			if (editables[--j].getId() === id) {
				return editables[j];
			}
		}

		return null;
	}

	/**
	 * Will initialize the contents that have been rendered in a given
	 * container for front end editing.
	 *
	 * @TODO: This function should be moved out of gcn library.  We should
	 *        publish a message instead, and pass these arguments in the
	 *        message.
	 *
	 * @private
	 * @static
	 * @param {Array.<object>} editables Editables to be `aloha()'fied.
	 * @param {Array.<object>} blocks Blocks to receive tagfill buttons.
	 * @param {number|string} pageId id of the page the tag belongs to
	 * @param {jQuery<HTMLElement>} container The element that wraps the
	 *                              incoming tag contents.
	 */
	function initializeFrontendEditing(editables, blocks, pageId, container) {
		var Aloha = (typeof window !== 'undefined') && window.Aloha;

		if (!Aloha) {
			return;
		}

		Aloha.ready(function () {
			if (Aloha.GCN) {
				// If we are in the backend, then we need to remove the id of
				// the container because it is duplicated in the incoming
				// content.
				if (container && Aloha.GCN.isBackendMode()) {
					container.removeAttr('id');
				}
				Aloha.GCN.page = GCN.page(pageId);
				Aloha.GCN.setupConstructsButton(pageId);
			}

			var j = editables && editables.length;
			var editable;
			var unmodified = [];
			while (j) {
				Aloha.jQuery('#' + editables[--j].element).aloha();
				editable = getAlohaEditableById(editables[j].element);
				if (editable) {
					unmodified.push(editable);
					if (editables[j].readonly) {
						editable.disable();
					}
				}
			}

			if (Aloha.GCN) {
				j = Aloha.editables.length;
				while (j) {
					if (!Aloha.editables[--j].isModified()) {
						unmodified.push(Aloha.editables[j]);
					}
				}
				Aloha.GCN.alohaBlocks(blocks, pageId, function () {
					var j = unmodified.length;
					while (j) {
						unmodified[--j].setUnmodified();
					}
				});
			}
		});
	}

	/**
	 * Helper function to normalize the arguments that can be passed to the
	 * `edit()' and `render()' methods.
	 *
	 * @private
	 * @static
	 * @param {arguments} args A list of arguments.
	 * @return {object} Object containing an the properties `element',
	 *                         `success' and `error'.
	 */
	function getRenderOptions(args) {
		var argv = Array.prototype.slice.call(args);
		var argc = args.length;
		var arg;
		var i;

		var element;
		var success;
		var error;

		for (i = 0; i < argc; ++i) {
			arg = argv[i];

			switch (jQuery.type(arg)) {
			case 'string':
				element = jQuery(arg);
				break;
			case 'object':
				element = arg;
				break;
			case 'function':
				if (success) {
					error = arg;
				} else {
					success = arg;
				}
				break;
			// Descarding all other types of arguments...
			}
		}

		return {
			element : element,
			success : success,
			error   : error
		};
	}

	/**
	 * Exposes an API to operate on a Content.Node tag.
	 *
	 * @public
	 * @class TagAPI
	 */
	var TagAPI = GCN.defineChainback({

		__chainbacktype__: 'TagAPI',

		/**
		 * @type {GCN.ContentObject} A reference to the object in which this
		 *                           tag is contained.  This value is set
		 *                           during initialization.
		 */
		_parent: null,

		/**
		 * @type {string} Name of this tag.
		 */
		_name: null,

		/**
		 * Gets this tag's information from the object that contains it.
		 *
		 * @param {function(TagAPI)} success Callback to be invoked when this
		 *                                   operation completes normally.
		 * @param {function(GCNError):boolean} error Custom error handler.
		 */
		'!_read': function (success, error) {
			if (this._fetched) {
				if (success) {
					success(this);
				}

				return;
			}

			var that = this;
			var parent = this.parent();

			// assert(parent)

			// Take the data for this tag from it's container.
			parent._read(function () {
				that._data = parent._getTagData(that._name);

				if (!that._data) {
					var err = GCN.createError('TAG_NOT_FOUND',
						'Could not find tag "' + that._name + '" in ' +
						parent._type + " " + parent._data.id, that);

					GCN.handleError(err, error);

					return;
				}

				that._fetched = true;

				if (success) {
					success(that);
				}
			}, error);
		},

		/**
		 * Retrieve the object in which this tag is contained.  It does so by
		 * getting this chainback's "chainlink ancestor" object.
		 *
		 * @return {GCN.AbstractTagContainer}
		 */
		'!parent': function () {
			return this._ancestor();
		},

		/**
		 * Initialize a tag object.  Unlike other chainback objects, tags will
		 * always have a parent.  If its parent have been loaded, we will
		 * immediately copy the this tag's data from the parent's `_data'
		 * object to the tag's `_data' object.
		 *
		 * @param {string|object} settings
		 * @param {function(TagAPI)} success Callback to be invoked when this
		 *                                   operation completes normally.
		 * @param {function(GCNError):boolean} error Custom error handler.
		 */
		_init: function (settings, success, error) {
			if (jQuery.type(settings) === 'object') {
				this._name    = settings.name;
				this._data    = settings;
				this._data.id = settings.id;
				this._fetched = true;
			} else {
				this._data = {};
				this._data.id = this._name = settings;
			}

			if (success) {
				var that = this;

				this._read(function (container) {
					that._read(success, error);
				}, error);

			// Even if not success callback is given, read this tag's data from
			// is container, it that container has the data available.
			// If we are initializing a placeholder tag object (in the process
			// of creating brand new tag, for example), then its parent
			// container will not have any data for this tag yet.  We know that
			// we are working with a placeholder tag if no `_data.id' or `_name'
			// property is set.
			} else if (!this._fetched && this._name &&
			           this.parent()._fetched) {
				this._data = this.parent()._getTagData(this._name);
				this._fetched = !!this._data;

			// We are propably initializing a placholder object, we will assign
			// it its own `_data' and `_fetched' properties so that it is not
			// accessing the prototype values.
			} else if (!this._fetched) {
				this._data = {};
				this._data.id = this._name = settings;
				this._fetched = false;
			}
		},

		/**
		 * Get or set a property of this tags.
		 * Note that tags do not have a `_shadow' object, and we update the
		 * `_data' object directly.
		 *
		 * @param {string} name Name of tag part.
		 * @param {*=} set Optional value.  If provided, the tag part will be
		 *                 replaced with this value.
		 * @return {*} The value of the accessed tag part.
		 * @throws UNFETCHED_OBJECT_ACCESS
		 */
		'!prop': function (name, value) {
			var parent = this.parent();

			if (!this._fetched) {
				GCN.error('UNFETCHED_OBJECT_ACCESS',
					'Calling method `prop()\' on an unfetched object: ' +
					parent._type + " " + parent._data.id, this);

				return;
			}

			if (jQuery.type(value) !== 'undefined') {
				this._data[name] = value;
				parent._update('tags.' + name, this._data);
			}

			return this._data[name];
		},

		/**
		 * Get or set a part of this tags.
		 *
		 * @param {string} name Name of tag opart.
		 * @param {*=} set Optional value.  If provided, the tag part will be
		 *                 replaced with this value.
		 * @return {*} The value of the accessed tag part.
		 * @throws UNFETCHED_OBJECT_ACCESS
		 * @throws PART_NOT_FOUND
		 */
		'!part': function (name, value) {
			var parent;

			if (!this._fetched) {
				parent = this.parent();

				GCN.error('UNFETCHED_OBJECT_ACCESS',
					'Calling method `prop()\' on an unfetched object: ' +
					parent._type + " " + parent._data.id, this);

				return null;
			}

			var part = this._data.properties[name];

			if (!part) {
				parent = this.parent();

				GCN.error('PART_NOT_FOUND', 'Tag "' + this._name +
					'" of ' + parent._type + ' ' + parent._data.id +
					' does not have a part "' + name + '"', this);

				return null;
			}

			if (jQuery.type(value) === 'undefined') {
				return getPartValue(part);
			}

			setPartValue(part, value);

			// Each time we perform a write operation on a tag, we will update
			// the tag in the tag container's `_shadow' object as well.
			this.parent()._update('tags.' + this._name, this._data);

			return value;
		},

		/**
		 * Remove this tag from its containing object (it's parent).
		 *
		 * @param {function} callback A function that receive this tag's parent
		 *                            object as its only arguments.
		 */
		remove: function (success, error) {
			var parent = this.parent();

			if (!parent.hasOwnProperty('_deletedTags')) {
				parent._deletedTags = [];
			}

			parent._deletedTags.push(this._name);

			if (parent._data.tags &&
					parent._data.tags[this._name]) {
				delete parent._data.tags[this._name];
			}

			if (parent._shadow.tags &&
					parent._shadow.tags[this._name]) {
				delete parent._shadow.tags[this._name];
			}

			if (success) {
				parent._persist(success, error);
			}
		},

		/**
		 * Will render this tag in the given render `mode'.  If an element is
		 * provided, the content will be placed in that element.  If the `mode'
		 * is "edit", any rendered editables will be initialized for Aloha
		 * Editor.  Any editable that are rendered into an element will also be
		 * added to the tag's parent object's `_editables' array so that they
		 * can have their changed contents copied back into their corresponding
		 * tags during saving.
		 *
		 * @param {string} mode The rendering mode.  Valid values are "view",
		 *                      and "edit".
		 * @param {jQuery.<HTMLElement>} element DOM element into which the
		 *                                       the rendered content should be
		 *                                       placed.
		 * @param {function(string, TagAPI, object)} Optional success handler.
		 * @param {function(GCNError):boolean} Optional custom error handler.
		 */
		'!_render': function (mode, element, success, error) {
			var that = this;
			var parent = this.parent();

			this._read(function () {
				var template = '<node ' + that._name + '>';

				that._procure();

				parent._renderTemplate(template, mode, function (data) {
					var tags = parent._getEditablesAndBlocks(data);

					parent._storeRenderedEditables(tags.editables);
					parent._storeRenderedBlocks(tags.blocks);

					GCN._handleContentRendered(data.content, function (html) {
						if (element) {
							element.html(html);
							GCN.pub('content-inserted', [element, html]);
						}

						if (success) {
							success(html, that, data);
						}

						initializeFrontendEditing(tags.editables, tags.blocks,
							parent.id(), element);

						that._vacate();
					});
				}, function () {
					that._vacate();
				});
			}, error);
		},

		/**
		 * Render the tag based on its settings on the server.
		 * Can be called with the following arguments:
		 *
		 * Do nothing:
		 * render()
		 *
		 * Render tag contents into div whose id is "content-div":
		 * @param {string|jQuery.<HTMLElement>}
		 * render('#content-div') or render(jQuery('#content-div'))
		 *
		 * Pass the html rendering of the tag in the given callback:
		 * @param {function(string, GCN.TagAPI)}
		 * render(function (html, tag) {})
		 *
		 * Whenever a 2nd argument is provided, it will be taken as as custom
		 * error handler.
		 */
		render: function () {
			var that = this;
			var args = arguments;

			// Wait until DOM is ready
			jQuery(function () {
				args = getRenderOptions(args);

				if (args.element || args.success) {
					that._render('view', args.element, args.success,
						args.error);
				}
			});
		},

		/**
		 * Like `render()', except that the content is rendered with additional
		 * elements that are required for front-end editing. ie: editables.
		 */
		edit: function () {
			var that = this;
			var args = arguments;

			// Wait until DOM is ready
			jQuery(function () {
				args = getRenderOptions(args);

				if (args.element || args.success) {
					that._render('edit', args.element, args.success,
						args.error);
				}
			});
		},

		/**
		 * Persists the changes to this tag on its container object.
		 *
		 * @param {function(TagAPI)} success Callback to be invoked when this
		 *                                   operation completes normally.
		 * @param {function(GCNError):boolean} error Custom error handler.
		 */
		save: function (success, error) {
			var that = this;
			this.parent().save(function () {
				if (success) {
					success(that);
				}
			}, error);
		}

	});

	// Unlike content objects, tags do not have unique ids and so we uniquely I
	// dentify tags by their name, and their parent's id.
	TagAPI._needsChainedHash = true;

	GCN.tag = GCN.exposeAPI(TagAPI);
	GCN.TagAPI = TagAPI;

}(GCN));
